@@ -5,18 +5,27 @@ This module contains the base classes for implementing handlers.
 
 from __future__ import annotations
 
+import datetime
 import importlib
+import inspect
 import sys
+from concurrent import futures
+from io import BytesIO
 from pathlib import Path
-from typing import TYPE_CHECKING, Any, BinaryIO, ClassVar, Iterable, Iterator, Mapping, MutableMapping, Sequence, cast
+from typing import TYPE_CHECKING, Any, BinaryIO, ClassVar, cast
+from warnings import warn
 from xml.etree.ElementTree import Element, tostring
 
 from jinja2 import Environment, FileSystemLoader
 from markdown import Markdown
 from markdown.extensions.toc import TocTreeprocessor
 from markupsafe import Markup
-from mkdocs_autorefs.references import AutorefsInlineProcessor
+from mkdocs_autorefs import AutorefsInlineProcessor
 
+# TODO: Replace with `from mkdocs.utils.cache import download_and_cache_url` when we depend on mkdocs>=1.5.
+from mkdocs_get_deps.cache import download_and_cache_url
+
+from mkdocstrings._download import download_url_with_gz
 from mkdocstrings.handlers.rendering import (
     HeadingShiftingTreeprocessor,
     Highlighter,
@@ -25,7 +34,7 @@ from mkdocstrings.handlers.rendering import (
     ParagraphStrippingTreeprocessor,
 )
 from mkdocstrings.inventory import Inventory
-from mkdocstrings.loggers import get_template_logger
+from mkdocstrings.loggers import get_logger, get_template_logger
 
 # TODO: remove once support for Python 3.9 is dropped
 if sys.version_info < (3, 10):
@@ -34,9 +43,25 @@ else:
     from importlib.metadata import entry_points
 
 if TYPE_CHECKING:
-    from mkdocs_autorefs.references import AutorefsHookInterface
+    from collections.abc import Iterable, Iterator, Mapping, Sequence
+
+    from markdown import Extension
+    from mkdocs_autorefs import AutorefsHookInterface
+
+log = get_logger(__name__)
 
 CollectorItem = Any
+HandlerConfig = Any
+HandlerOptions = Any
+
+
+# Autodoc instructions can appear in nested Markdown,
+# so we need to keep track of the Markdown conversion layer we're in.
+# Since any handler can be called from any Markdown conversion layer,
+# we need to keep track of the layer globally.
+# This global variable is incremented/decremented in `do_convert_markdown`,
+# and used in `outer_layer`.
+_markdown_conversion_layer: int = 0
 
 
 class CollectionError(Exception):
@@ -78,44 +103,123 @@ class BaseHandler:
     To add custom CSS, add an `extra_css` variable or create an 'style.css' file beside the templates.
     """
 
-    # TODO: Make name mandatory?
-    name: str = ""
+    # YORE: Bump 1: Replace ` = ""` with `` within line.
+    name: ClassVar[str] = ""
     """The handler's name, for example "python"."""
-    domain: str = "default"
+
+    # YORE: Bump 1: Replace ` = ""` with `` within line.
+    domain: ClassVar[str] = ""
     """The handler's domain, used to register objects in the inventory, for example "py"."""
-    enable_inventory: bool = False
+
+    enable_inventory: ClassVar[bool] = False
     """Whether the inventory creation is enabled."""
+
+    # YORE: Bump 1: Remove block.
     fallback_config: ClassVar[dict] = {}
     """Fallback configuration when searching anchors for identifiers."""
-    fallback_theme: str = ""
+
+    fallback_theme: ClassVar[str] = ""
     """Fallback theme to use when a template isn't found in the configured theme."""
-    extra_css = ""
+
+    extra_css: str = ""
     """Extra CSS."""
 
-    def __init__(self, handler: str, theme: str, custom_templates: str | None = None) -> None:
+    def __init__(
+        self,
+        # YORE: Bump 1: Remove line.
+        *args: Any,
+        # YORE: Bump 1: Remove line.
+        **kwargs: Any,
+        # YORE: Bump 1: Replace `# ` with `` within block.
+        # *,
+        # theme: str,
+        # custom_templates: str | None,
+        # mdx: Sequence[str | Extension],
+        # mdx_config: Mapping[str, Any],
+    ) -> None:
         """Initialize the object.
 
         If the given theme is not supported (it does not exist), it will look for a `fallback_theme` attribute
         in `self` to use as a fallback theme.
 
-        Arguments:
-            handler: The name of the handler.
-            theme: The name of theme to use.
-            custom_templates: Directory containing custom templates.
+        Keyword Arguments:
+            theme (str): The theme to use.
+            custom_templates (str | None): The path to custom templates.
+            mdx (list[str | Extension]): A list of Markdown extensions to use.
+            mdx_config (Mapping[str, Mapping[str, Any]]): Configuration for the Markdown extensions.
         """
+        # YORE: Bump 1: Remove block.
+        handler = ""
+        theme = ""
+        custom_templates = None
+        if args:
+            handler, args = args[0], args[1:]
+        if args:
+            theme, args = args[0], args[1:]
+            warn(
+                "The `theme` argument must be passed as a keyword argument.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        if args:
+            custom_templates, args = args[0], args[1:]
+            warn(
+                "The `custom_templates` argument must be passed as a keyword argument.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        handler = kwargs.pop("handler", handler)
+        theme = kwargs.pop("theme", theme)
+        custom_templates = kwargs.pop("custom_templates", custom_templates)
+        mdx = kwargs.pop("mdx", None)
+        mdx_config = kwargs.pop("mdx_config", None)
+        if handler:
+            if not self.name:
+                type(self).name = handler
+            warn(
+                "The `handler` argument is deprecated. The handler name must be specified as a class attribute.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        if not self.domain:
+            warn(
+                "The `domain` attribute must be specified as a class attribute.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        if mdx is None:
+            warn(
+                "The `mdx` argument must be provided (as a keyword argument).",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        if mdx_config is None:
+            warn(
+                "The `mdx_config` argument must be provided (as a keyword argument).",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+
+        self.theme = theme
+        self.custom_templates = custom_templates
+        self.mdx = mdx
+        self.mdx_config = mdx_config
+        self._md: Markdown | None = None
+        self._headings: list[Element] = []
+
         paths = []
 
         # add selected theme templates
-        themes_dir = self.get_templates_dir(handler)
-        paths.append(themes_dir / theme)
+        themes_dir = self.get_templates_dir(self.name)
+        paths.append(themes_dir / self.theme)
 
         # add extended theme templates
-        extended_templates_dirs = self.get_extended_templates_dirs(handler)
+        extended_templates_dirs = self.get_extended_templates_dirs(self.name)
         for templates_dir in extended_templates_dirs:
-            paths.append(templates_dir / theme)
+            paths.append(templates_dir / self.theme)
 
         # add fallback theme templates
-        if self.fallback_theme and self.fallback_theme != theme:
+        if self.fallback_theme and self.fallback_theme != self.theme:
             paths.append(themes_dir / self.fallback_theme)
 
             # add fallback theme of extended templates
@@ -128,19 +232,33 @@ class BaseHandler:
                 self.extra_css += "\n" + css_path.read_text(encoding="utf-8")
                 break
 
-        if custom_templates is not None:
-            paths.insert(0, Path(custom_templates) / handler / theme)
+        if self.custom_templates is not None:
+            paths.insert(0, Path(self.custom_templates) / self.name / self.theme)
 
         self.env = Environment(
             autoescape=True,
             loader=FileSystemLoader(paths),
             auto_reload=False,  # Editing a template in the middle of a build is not useful.
         )
+        self.env.filters["convert_markdown"] = self.do_convert_markdown
+        self.env.filters["heading"] = self.do_heading
         self.env.filters["any"] = do_any
         self.env.globals["log"] = get_template_logger(self.name)
 
-        self._headings: list[Element] = []
-        self._md: Markdown = None  # type: ignore[assignment]  # To be populated in `update_env`.
+    @property
+    def md(self) -> Markdown:
+        """The Markdown instance.
+
+        Raises:
+            RuntimeError: When the Markdown instance is not set yet.
+        """
+        if self._md is None:
+            raise RuntimeError("Markdown instance not set yet")
+        return self._md
+
+    def get_inventory_urls(self) -> list[tuple[str, dict[str, Any]]]:
+        """Return the URLs (and configuration options) of the inventory files to download."""
+        return []
 
     @classmethod
     def load_inventory(
@@ -163,7 +281,22 @@ class BaseHandler:
         """
         yield from ()
 
-    def collect(self, identifier: str, config: MutableMapping[str, Any]) -> CollectorItem:
+    def get_options(self, local_options: Mapping[str, Any]) -> HandlerOptions:
+        """Get combined options.
+
+        Override this method to customize how options are combined,
+        for example by merging the global options with the local options.
+        By combining options here, you don't have to do it twice in `collect` and `render`.
+
+        Arguments:
+            local_options: The local options.
+
+        Returns:
+            The combined options.
+        """
+        return local_options
+
+    def collect(self, identifier: str, options: HandlerOptions) -> CollectorItem:
         """Collect data given an identifier and user configuration.
 
         In the implementation, you typically call a subprocess that returns JSON, and load that JSON again into
@@ -173,19 +306,19 @@ class BaseHandler:
             identifier: An identifier for which to collect data. For example, in Python,
                 it would be 'mkdocstrings.handlers' to collect documentation about the handlers module.
                 It can be anything that you can feed to the tool of your choice.
-            config: The handler's configuration options.
+            options: The final configuration options.
 
         Returns:
             Anything you want, as long as you can feed it to the handler's `render` method.
         """
         raise NotImplementedError
 
-    def render(self, data: CollectorItem, config: Mapping[str, Any]) -> str:
+    def render(self, data: CollectorItem, options: HandlerOptions) -> str:
         """Render a template using provided data and configuration options.
 
         Arguments:
             data: The collected data to render.
-            config: The handler's configuration options.
+            options: The final configuration options.
 
         Returns:
             The rendered template as HTML.
@@ -239,17 +372,22 @@ class BaseHandler:
         discovered_extensions = entry_points(group=f"mkdocstrings.{handler}.templates")
         return [extension.load()() for extension in discovered_extensions]
 
-    def get_anchors(self, data: CollectorItem) -> tuple[str, ...]:  # noqa: ARG002
-        """Return the possible identifiers (HTML anchors) for a collected item.
+    def get_aliases(self, identifier: str) -> tuple[str, ...]:  # noqa: ARG002
+        """Return the possible aliases for a given identifier.
 
         Arguments:
-            data: The collected data.
+            identifier: The identifier to get the aliases of.
 
         Returns:
-            The HTML anchors (without '#'), or an empty tuple if this item doesn't have an anchor.
+            A tuple of strings - aliases.
         """
         return ()
 
+    @property
+    def outer_layer(self) -> bool:
+        """Whether we're in the outer Markdown conversion layer."""
+        return _markdown_conversion_layer == 0
+
     def do_convert_markdown(
         self,
         text: str,
@@ -270,22 +408,25 @@ class BaseHandler:
         Returns:
             An HTML string.
         """
-        treeprocessors = self._md.treeprocessors
+        global _markdown_conversion_layer  # noqa: PLW0603
+        _markdown_conversion_layer += 1
+        treeprocessors = self.md.treeprocessors
         treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = heading_level  # type: ignore[attr-defined]
         treeprocessors[IdPrependingTreeprocessor.name].id_prefix = html_id and html_id + "--"  # type: ignore[attr-defined]
         treeprocessors[ParagraphStrippingTreeprocessor.name].strip = strip_paragraph  # type: ignore[attr-defined]
 
         if autoref_hook:
-            self._md.inlinePatterns[AutorefsInlineProcessor.name].hook = autoref_hook  # type: ignore[attr-defined]
+            self.md.inlinePatterns[AutorefsInlineProcessor.name].hook = autoref_hook  # type: ignore[attr-defined]
 
         try:
-            return Markup(self._md.convert(text))
+            return Markup(self.md.convert(text))
         finally:
             treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = 0  # type: ignore[attr-defined]
             treeprocessors[IdPrependingTreeprocessor.name].id_prefix = ""  # type: ignore[attr-defined]
             treeprocessors[ParagraphStrippingTreeprocessor.name].strip = False  # type: ignore[attr-defined]
-            self._md.inlinePatterns[AutorefsInlineProcessor.name].hook = None  # type: ignore[attr-defined]
-            self._md.reset()
+            self.md.inlinePatterns[AutorefsInlineProcessor.name].hook = None  # type: ignore[attr-defined]
+            self.md.reset()
+            _markdown_conversion_layer -= 1
 
     def do_heading(
         self,
@@ -336,7 +477,7 @@ class BaseHandler:
         el = Element(f"h{heading_level}", attributes)
         el.append(Element("mkdocstrings-placeholder"))
         # Tell the inner 'toc' extension to make its additions if configured so.
-        toc = cast(TocTreeprocessor, self._md.treeprocessors["toc"])
+        toc = cast(TocTreeprocessor, self.md.treeprocessors["toc"])
         if toc.use_anchors:
             toc.add_anchor(el, attributes["id"])
         if toc.use_permalinks:
@@ -362,29 +503,45 @@ class BaseHandler:
         self._headings.clear()
         return result
 
-    def update_env(self, md: Markdown, config: dict) -> None:  # noqa: ARG002
-        """Update the Jinja environment.
+    # YORE: Bump 1: Replace `*args: Any, **kwargs: Any` with `config: Any`.
+    def update_env(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
+        """Update the Jinja environment."""
+        # YORE: Bump 1: Remove line.
+        warn("No need to call `super().update_env()` anymore.", DeprecationWarning, stacklevel=2)
 
-        Arguments:
-            md: The Markdown instance. Useful to add functions able to convert Markdown into the environment filters.
-            config: Configuration options for `mkdocs` and `mkdocstrings`, read from `mkdocs.yml`. See the source code
-                of [mkdocstrings.plugin.MkdocstringsPlugin.on_config][] to see what's in this dictionary.
-        """
-        self._md = md
-        self.env.filters["highlight"] = Highlighter(md).highlight
-        self.env.filters["convert_markdown"] = self.do_convert_markdown
-        self.env.filters["heading"] = self.do_heading
-
-    def _update_env(self, md: Markdown, config: dict) -> None:
+    def _update_env(self, md: Markdown, *, config: Any | None = None) -> None:
         """Update our handler to point to our configured Markdown instance, grabbing some of the config from `md`."""
-        extensions = config["mdx"] + [MkdocstringsInnerExtension(self._headings)]
+        # YORE: Bump 1: Remove block.
+        if self.mdx is None and config is not None:
+            self.mdx = config.get("mdx", None) or config.get("markdown_extensions", None) or ()
+        if self.mdx_config is None and config is not None:
+            self.mdx_config = config.get("mdx_config", None) or config.get("mdx_configs", None) or {}
+
+        extensions: list[str | Extension] = [*self.mdx, MkdocstringsInnerExtension(self._headings)]
+
+        new_md = Markdown(extensions=extensions, extension_configs=self.mdx_config)
 
-        new_md = Markdown(extensions=extensions, extension_configs=config["mdx_configs"])
         # MkDocs adds its own (required) extension that's not part of the config. Propagate it.
         if "relpath" in md.treeprocessors:
-            new_md.treeprocessors.register(md.treeprocessors["relpath"], "relpath", priority=0)
+            relpath = md.treeprocessors["relpath"]
+            new_relpath = type(relpath)(relpath.file, relpath.files, relpath.config)  # type: ignore[attr-defined,call-arg]
+            new_md.treeprocessors.register(new_relpath, "relpath", priority=0)
 
-        self.update_env(new_md, config)
+        self._md = new_md
+
+        self.env.filters["highlight"] = Highlighter(new_md).highlight
+
+        # YORE: Bump 1: Replace block with `self.update_env(config)`.
+        parameters = inspect.signature(self.update_env).parameters
+        if "md" in parameters:
+            warn(
+                "The `update_env(md)` parameter is deprecated. Use `self.md` instead.",
+                DeprecationWarning,
+                stacklevel=1,
+            )
+            self.update_env(new_md, config)
+        elif "config" in parameters:
+            self.update_env(config)
 
 
 class Handlers:
@@ -394,17 +551,46 @@ class Handlers:
     this for the purpose of caching. Use [mkdocstrings.plugin.MkdocstringsPlugin.get_handler][] for convenient access.
     """
 
-    def __init__(self, config: dict) -> None:
+    def __init__(
+        self,
+        *,
+        theme: str,
+        default: str,
+        inventory_project: str,
+        inventory_version: str = "0.0.0",
+        handlers_config: dict[str, HandlerConfig] | None = None,
+        custom_templates: str | None = None,
+        mdx: Sequence[str | Extension] | None = None,
+        mdx_config: Mapping[str, Any] | None = None,
+        tool_config: Any,
+    ) -> None:
         """Initialize the object.
 
         Arguments:
-            config: Configuration options for `mkdocs` and `mkdocstrings`, read from `mkdocs.yml`. See the source code
-                of [mkdocstrings.plugin.MkdocstringsPlugin.on_config][] to see what's in this dictionary.
+            theme: The theme to use.
+            default: The default handler to use.
+            inventory_project: The project name to use in the inventory.
+            inventory_version: The project version to use in the inventory.
+            handlers_config: The handlers configuration.
+            custom_templates: The path to custom templates.
+            mdx: A list of Markdown extensions to use.
+            mdx_config: Configuration for the Markdown extensions.
+            tool_config: Tool configuration to pass down to handlers.
         """
-        self._config = config
+        self._theme = theme
+        self._default = default
+        self._handlers_config = handlers_config or {}
+        self._custom_templates = custom_templates
+        self._mdx = mdx or []
+        self._mdx_config = mdx_config or {}
         self._handlers: dict[str, BaseHandler] = {}
-        self.inventory: Inventory = Inventory(project=self._config["mkdocs"]["site_name"])
+        self._tool_config = tool_config
 
+        self.inventory: Inventory = Inventory(project=inventory_project, version=inventory_version)
+
+        self._inv_futures: dict[futures.Future, tuple[BaseHandler, str, Any]] = {}
+
+    # YORE: Bump 1: Remove block.
     def get_anchors(self, identifier: str) -> tuple[str, ...]:
         """Return the canonical HTML anchor for the identifier, if any of the seen handlers can collect it.
 
@@ -415,13 +601,22 @@ class Handlers:
             A tuple of strings - anchors without '#', or an empty tuple if there isn't any identifier familiar with it.
         """
         for handler in self._handlers.values():
-            fallback_config = getattr(handler, "fallback_config", {})
             try:
-                anchors = handler.get_anchors(handler.collect(identifier, fallback_config))
+                if hasattr(handler, "get_anchors"):
+                    warn(
+                        "The `get_anchors` method is deprecated. "
+                        "Declare a `get_aliases` method instead, accepting a string (identifier) "
+                        "instead of a collected object.",
+                        DeprecationWarning,
+                        stacklevel=1,
+                    )
+                    aliases = handler.get_anchors(handler.collect(identifier, getattr(handler, "fallback_config", {})))
+                else:
+                    aliases = handler.get_aliases(identifier)
             except CollectionError:
                 continue
-            if anchors:
-                return anchors
+            if aliases:
+                return aliases
         return ()
 
     def get_handler_name(self, config: dict) -> str:
@@ -433,10 +628,7 @@ class Handlers:
         Returns:
             The name of the handler to use.
         """
-        global_config = self._config["mkdocstrings"]
-        if "handler" in config:
-            return config["handler"]
-        return global_config["default_handler"]
+        return config.get("handler", self._default)
 
     def get_handler_config(self, name: str) -> dict:
         """Return the global configuration of the given handler.
@@ -447,10 +639,7 @@ class Handlers:
         Returns:
             The global configuration of the given handler. It can be an empty dictionary.
         """
-        handlers = self._config["mkdocstrings"].get("handlers", {})
-        if handlers:
-            return handlers.get(name, {})
-        return {}
+        return self._handlers_config.get(name, None) or {}
 
     def get_handler(self, name: str, handler_config: dict | None = None) -> BaseHandler:
         """Get a handler thanks to its name.
@@ -470,17 +659,91 @@ class Handlers:
         """
         if name not in self._handlers:
             if handler_config is None:
-                handler_config = self.get_handler_config(name)
-            handler_config.update(self._config)
+                handler_config = self._handlers_config.get(name, {})
             module = importlib.import_module(f"mkdocstrings_handlers.{name}")
-            self._handlers[name] = module.get_handler(
-                theme=self._config["theme_name"],
-                custom_templates=self._config["mkdocstrings"]["custom_templates"],
-                config_file_path=self._config["mkdocs"]["config_file_path"],
-                **handler_config,
-            )
+
+            # YORE: Bump 1: Remove block.
+            kwargs = {
+                "theme": self._theme,
+                "custom_templates": self._custom_templates,
+                "mdx": self._mdx,
+                "mdx_config": self._mdx_config,
+                "handler_config": handler_config,
+                "tool_config": self._tool_config,
+            }
+            if "config_file_path" in inspect.signature(module.get_handler).parameters:
+                kwargs["config_file_path"] = self._tool_config.get("config_file_path")
+                warn(
+                    "The `config_file_path` argument in `get_handler` functions is deprecated. "
+                    "Use `tool_config.get('config_file_path')` instead.",
+                    DeprecationWarning,
+                    stacklevel=1,
+                )
+            self._handlers[name] = module.get_handler(**kwargs)
+
+            # YORE: Bump 1: Replace `# ` with `` within block.
+            # self._handlers[name] = module.get_handler(
+            #     theme=self._theme,
+            #     custom_templates=self._custom_templates,
+            #     mdx=self._mdx,
+            #     mdx_config=self._mdx_config,
+            #     handler_config=handler_config,
+            #     tool_config=self._tool_config,
+            # )
         return self._handlers[name]
 
+    def _download_inventories(self) -> None:
+        """Download an inventory file from an URL.
+
+        Arguments:
+            url: The URL of the inventory.
+        """
+        to_download: list[tuple[BaseHandler, str, Any]] = []
+
+        for handler_name, conf in self._handlers_config.items():
+            handler = self.get_handler(handler_name)
+
+            if handler.get_inventory_urls.__func__ is BaseHandler.get_inventory_urls:  # type: ignore[attr-defined]
+                if inv_configs := conf.pop("import", ()):
+                    warn(
+                        "mkdocstrings v1 will stop handling 'import' in handlers configuration. "
+                        "Instead your handler must define a `get_inventory_urls` method "
+                        "that returns a list of URLs to download. ",
+                        DeprecationWarning,
+                        stacklevel=1,
+                    )
+                    inv_configs = [{"url": inv} if isinstance(inv, str) else inv for inv in inv_configs]
+                    inv_configs = [(inv.pop("url"), inv) for inv in inv_configs]
+            else:
+                inv_configs = handler.get_inventory_urls()
+
+            to_download.extend((handler, url, conf) for url, conf in inv_configs)
+
+        if to_download:
+            thread_pool = futures.ThreadPoolExecutor(4)
+            for handler, url, conf in to_download:
+                log.debug("Downloading inventory from %s", url)
+                future = thread_pool.submit(
+                    download_and_cache_url,
+                    url,
+                    datetime.timedelta(days=1),
+                    download=download_url_with_gz,
+                )
+                self._inv_futures[future] = (handler, url, conf)
+            thread_pool.shutdown(wait=False)
+
+    def _yield_inventory_items(self) -> Iterator[tuple[str, str]]:
+        if self._inv_futures:
+            log.debug("Waiting for %s inventory download(s)", len(self._inv_futures))
+            futures.wait(self._inv_futures, timeout=30)
+            # Reversed order so that pages from first futures take precedence:
+            for fut, (handler, url, conf) in reversed(self._inv_futures.items()):
+                try:
+                    yield from handler.load_inventory(BytesIO(fut.result()), url, **conf)
+                except Exception as error:  # noqa: BLE001
+                    log.error("Couldn't load inventory %s through handler '%s': %s", url, handler.name, error)  # noqa: TRY400
+            self._inv_futures = {}
+
     @property
     def seen_handlers(self) -> Iterable[BaseHandler]:
         """Get the handlers that were encountered so far throughout the build.
@@ -493,6 +756,8 @@ class Handlers:
 
     def teardown(self) -> None:
         """Teardown all cached handlers and clear the cache."""
+        for future in self._inv_futures:
+            future.cancel()
         for handler in self.seen_handlers:
             handler.teardown()
         self._handlers.clear()
